'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var expressionExpander = require('expression-expander');
var githubRepositoryProvider = require('github-repository-provider');

class Context {
  constructor(targetRepo, templateRepo, properties) {
    this.ctx = expressionExpander.createContext({
      keepUndefinedValues: true,
      leftMarker: '{{',
      rightMarker: '}}',
      markerRegexp: '{{([^}]+)}}'
    });

    this.ctx.properties = properties;

    Object.defineProperty(this, 'properties', {
      value: properties
    });

    Object.defineProperty(this, 'files', {
      value: new Map()
    });

    Object.defineProperty(this, 'targetRepo', {
      value: targetRepo
    });
    Object.defineProperty(this, 'templateRepo', {
      value: templateRepo,
      writable: true
    });

    this.logger = console;
  }

  expand(...args) {
    return this.ctx.expand(...args);
  }

  addFile(file) {
    this.files.set(file.path, file);
  }

  async usedDevModules() {
    const usedModuleSets = await Promise.all(
      Array.from(this.files.values()).map(async file => {
        if (file.path === 'package.json') {
          return file.usedDevModules(
            file.originalContent(this, { ignoreMissing: true })
          );
        } else {
          const m = await file.merge(this);
          return file.usedDevModules(m.content);
        }
      })
    );

    return usedModuleSets.reduce(
      (sum, current) => new Set([...sum, ...current]),
      new Set()
    );
  }

  optionalDevModules(modules) {
    return Array.from(this.files.values())
      .map(file => file.optionalDevModules(modules))
      .reduce((sum, current) => new Set([...sum, ...current]), new Set());
  }

  set text(value) {
    if (this.spinner === undefined) {
      console.log(value);
    } else {
      this.spinner.text = value;
    }
  }

  succeed(...args) {
    if (this.spinner === undefined) {
      console.log(...args);
    } else {
      this.spinner.succeed(...args);
    }
  }

  warn(...args) {
    if (this.spinner === undefined) {
      console.log(...args);
    } else {
      this.spinner.warn(...args);
    }
  }

  fail(...args) {
    if (this.spinner === undefined) {
      console.log(...args);
    } else {
      this.spinner.fail(...args);
    }
  }
}

class File {
  static matchesFileName(name) {
    return false;
  }

  constructor(path, options = {}) {
    Object.defineProperty(this, 'path', {
      value: path
    });

    Object.defineProperty(this, 'options', {
      value: options
    });
  }

  get needsTemplate() {
    return true;
  }

  get needsOriginal() {
    return false;
  }

  optionalDevModules(modules) {
    return new Set();
  }

  usedDevModules(content) {
    return new Set();
  }

  templateContent(context, options) {
    return this.content(context.templateRepo, this.path, options);
  }

  originalContent(context, options) {
    return this.content(context.targetRepo, this.path, options);
  }

  async mergeContent(context, original, template) {
    return {
      changed: false,
      content: original
    };
  }

  async merge(context) {
    const [original, template] = await Promise.all([
      this.originalContent(context, {
        ignoreMissing: !this.needsOriginal
      }),
      this.templateContent(context, {
        ignoreMissing: !this.needsTemplate
      })
    ]);

    return this.mergeContent(context, original, template);
  }

  async saveMerge(context) {
    try {
      const result = await this.merge(context);
      if (result === undefined) {
        return {
          path: this.path,
          changed: false
        };
      }
      result.path = this.path;

      context.properties.path = this.path;
      result.messages = context.expand(result.messages);

      return result;
    } catch (err) {
      context.fail(`${context.targetRepo.name},${this.path}: ${err}`);
      return {
        path: this.path,
        changed: false
      };
    }
  }

  async content(repository, path, options) {
    return repository.content(path, options);
  }
}

const yaml = require('js-yaml');
const deepExtend = require('deep-extend');
const semverDiff = require('semver-diff');

function diffVersion(a, b) {
  const aa = String(a)
    .split(/\./)
    .map(x => parseInt(x, 10));
  const bb = String(b)
    .split(/\./)
    .map(x => parseInt(x, 10));

  for (const i in aa) {
    if (i >= bb.length) {
      break;
    }

    if (aa[i] < bb[i]) {
      return -1;
    }
    if (aa[i] > bb[i]) {
      return 1;
    }
  }

  return 0;
}

function isValidVersion(version) {
  version = String(version);

  const m = version.match(/^(\d+)(\.(\d+))*$/);

  return m ? true : false;
}

function normalizeVersion(version) {
  version = String(version);

  const m = version.match(/^(\d+)(\.(\d+))?$/);

  if (m) {
    return m[3] ? m[1] + m[2] + '.0' : m[1] + '.0.0';
  }

  return version;
}

class Travis extends File {
  static matchesFileName(name) {
    return name === '.travis.yml';
  }

  async mergeContent(context, original, template) {
    const yml = yaml.safeLoad(original) || {};
    const tyml = yaml.safeLoad(context.expand(template));
    const before_script = yml.before_script;
    const email = yml.notifications ? yml.notifications.email : undefined;
    const formerNodeVersions = yml.node_js;
    const messages = [];

    let removeVersions = [];

    if (tyml.node_js) {
      removeVersions = tyml.node_js.filter(v => v < 0).map(v => -v);
      tyml.node_js = tyml.node_js.filter(v => (v < 0 ? false : true));
    }

    deepExtend(yml, tyml);

    if (formerNodeVersions !== undefined) {
      formerNodeVersions.forEach(ov => {
        if (
          isValidVersion(ov) &&
          yml.node_js.find(
            nv =>
              semverDiff(normalizeVersion(ov), normalizeVersion(nv)) === 'major'
          )
        ) {
          yml.node_js.push(ov);
        }
      });

      const toBeRemoved = yml.node_js.filter(v =>
        removeVersions.find(rv => diffVersion(rv, v) === 0)
      );
      if (toBeRemoved.length > 0) {
        messages.push(
          `chore(travis): remove node version(s) ${toBeRemoved.join(' ')}`
        );
      }

      yml.node_js = yml.node_js.filter(
        v => !removeVersions.find(rv => diffVersion(rv, v) === 0)
      );
    }

    if (email !== undefined) {
      yml.notifications.email = email;
    }

    if (before_script !== undefined) {
      before_script.forEach(s => {
        if (!yml.before_script.find(e => e === s)) {
          yml.before_script.push(s);
        }
      });
    }

    if (yml.before_script) {
      yml.before_script = yml.before_script.filter(
        s =>
          !tyml.before_script.find(e => e === `-${s}`) && s.indexOf('-') !== 0
      );
    }

    const content = yaml.safeDump(yml, {
      lineWidth: 128
    });

    if (messages.length === 0) {
      messages.push(`chore(travis): merge from template ${this.path}`);
    }

    return {
      content,
      changed: content !== original,
      messages
    };
  }
}

class Readme extends File {
  static matchesFileName(name) {
    return name.match(/README\./);
  }

  async mergeContent(context, original, template) {
    const pkgTemplate = await context.files
      .get('package.json')
      .templateContent(context);
    const pkg = await context.files
      .get('package.json')
      .originalContent(context);

    const p = JSON.parse(pkg);
    const pTemplate = JSON.parse(pkgTemplate);

    const badges =
      pTemplate.template && pTemplate.template.badges
        ? pTemplate.template.badges
            .map(b => {
              // TODO do not alter global properties use private layer here
              if (p.template !== undefined && p.template.badges !== undefined) {
                Object.assign(context.properties, p.template.badges[b.name]);
              }

              const r = context.expand(`[![${b.name}](${b.icon})](${b.url})`);

              if (r.match(/\{\{/)) {
                return '';
              }
              return r;
            })
            .filter(b => b.length > 0)
        : [];

    let body = original.split(/\n/);

    if (body.length === 0) {
      body = context.expand(template).split(/\n/);
    } else {
      body = body.filter(l => !l.match(/^\[\!\[.*\)$/));
    }

    const content = [...badges, ...body].join('\n');
    return {
      content,
      changed: content !== original,
      messages: ['docs(README): update from template']
    };
  }
}

function moduleNames(object) {
  if (object === undefined) return new Set();

  const modules = new Set();

  Object.keys(object).forEach(k => {
    const v = object[k];
    if (typeof v === 'string') {
      modules.add(v);
    } else if (Array.isArray(v)) {
      v.forEach(e => {
        if (typeof e === 'string') {
          modules.add(e);
        }
      });
    }
  });

  return modules;
}

class Package extends File {
  static matchesFileName(name) {
    return name.match(/^package\.json$/);
  }

  optionalDevModules(modules = new Set()) {
    return new Set(['cracks', 'dont-crack'].filter(m => modules.has(m)));
  }

  async usedDevModules(content) {
    content = await content;

    const pkg = content.length === 0 ? {} : JSON.parse(content);

    return moduleNames(pkg.release);
  }

  async templateRepo(context) {
    const content = await this.originalContent(context, {
      ignoreMissing: true
    });
    const pkg = JSON.parse(content);
    if (pkg.template !== undefined && pkg.template.repository !== undefined) {
      const m = pkg.template.repository.url.match(/github.com\/(.*)\.git$/);
      return m[1];
    }

    return undefined;
  }

  async mergeContent(context, original, templateContent) {
    const originalLastChar = original[original.length - 1];

    let target =
      original === undefined || original === '' ? {} : JSON.parse(original);
    const template = JSON.parse(templateContent);
    let messages = [];
    const properties = context.properties;

    if (target.name === undefined || target.name === '') {
      const m = context.targetRepo.name.match(/^([^\/]+)\/(.*)/);
      target.name = m ? m[2] : context.targetRepo.name;
    }

    if (target.module !== undefined && !target.module.match(/\{\{module\}\}/)) {
      properties.module = target.module;
    }

    properties.main =
      target.main && !target.main.match(/\{\{main\}\}/)
        ? target.main
        : 'dist/index.js';

    const githubURL = `git+https://github.com/${properties.user}/${
      target.name
    }.git`;
    const githubURLAlternative = `git+https://github.com/${
      properties.user
    }/node-${target.name}.git`;

    let repoName = target.name;

    if (
      target.repository === undefined ||
      (target.repository.url !== githubURL &&
        target.repository.url !== githubURLAlternative)
    ) {
      target.repository = {
        type: 'git',
        url: githubURL
      };
      messages.push(`chore(package): correct github url`);
    }

    if (
      target.repository !== undefined &&
      githubURLAlternative === target.repository.url
    ) {
      repoName = `node-${target.name}`;
    }

    const bugsURL = `https://github.com/${properties.user}/${repoName}/issues`;

    if (target.bugs === undefined || target.bugs.url !== bugsURL) {
      target.bugs = {
        url: bugsURL
      };
      messages.push(`chore(package): correct bugs url`);
    }

    const homepageURL = `https://github.com/${properties.user}/${
      repoName
    }#readme`;

    if (target.homepage === undefined || target.homepage !== homepageURL) {
      target.homepage = homepageURL;
      messages.push(`chore(package): correct hompage url`);
    }

    let buildOutput;
    const extraBuilds = {};

    if (target.scripts !== undefined) {
      Object.keys(target.scripts).forEach(key => {
        const script = target.scripts[key];
        const ma = script.match(/--output=([^\s]+)/);
        if (ma) {
          buildOutput = ma[1];
        }

        if (template.scripts !== undefined && key !== 'semantic-release') {
          const templateScript = template.scripts[key];
          if (templateScript !== undefined) {
            const mb = script.match(/&&\s*(.+)$/);
            if (mb && !templateScript.includes(mb[1])) {
              extraBuilds[key] = mb[1];
            }
          }
        }
      });
    }

    const usedDevModules = await context.usedDevModules();

    context.logger.debug(
      `usedDevModules: ${Array.from(usedDevModules).join(',')}`
    );

    const optionalDevModules = context.optionalDevModules(usedDevModules);

    context.logger.debug(
      `optionalDevModules: ${Array.from(optionalDevModules).join(',')}`
    );

    const deepProperties = {
      devDependencies: {},
      dependencies: {},
      peerDependencies: {},
      optionalDependencies: {},
      scripts: {},
      engines: {}
    };

    Object.keys(deepProperties).forEach(category => {
      if (template[category] !== undefined) {
        Object.keys(template[category]).forEach(d => {
          if (template[category][d] === '-') {
            if (
              target[category] !== undefined &&
              target[category][d] !== undefined
            ) {
              messages.push(
                `chore(${category}): remove ${d}@${target[category][d]}`
              );
              delete target[category][d];
            }
          } else {
            const tp = context.expand(template[category][d]);

            if (
              category === 'devDependencies' &&
              target.dependencies !== undefined &&
              target.dependencies[d] === tp
            ) {
              // do not include dev dependency if regular dependency is already present
            } else {
              if (
                target[category] !== undefined &&
                tp !== target[category][d]
              ) {
                messages.push(
                  target[category][d] === undefined
                    ? `chore(${category}): add ${d}@${tp} from template`
                    : `chore(${category}): update ${d}@${tp} from template`
                );
                target[category][d] = tp;
              }
            }
          }
        });
      }
    });

    Object.keys(template).forEach(p => {
      if (p !== 'template') {
        if (target[p] === undefined) {
          target[p] = template[p];
        }
      }
    });

    if (target.scripts !== undefined && target.scripts.prepare) {
      if (buildOutput !== undefined) {
        target.scripts.prepare = target.scripts.prepare.replace(
          /--output=([^\s]+)/,
          `--output=${buildOutput}`
        );
      }
    }

    Object.keys(extraBuilds).forEach(key => {
      target.scripts[key] += ` && ${extraBuilds[key]}`;
    });

    if (target.module === '{{module}}') {
      delete target.module;
    }

    if (
      target.contributors !== undefined &&
      target.author !== undefined &&
      target.author.name !== undefined
    ) {
      const m = target.author.name.match(/(^[^<]+)<([^>]+)>/);
      if (m !== undefined) {
        const name = String(m[1]).replace(/^\s+|\s+$/g, '');
        const email = m[2];

        if (
          target.contributors.find(c => c.name === name && c.email === email)
        ) {
          delete target.author;
        }
      }
    }

    if (
      target.template === undefined ||
      target.template.repository === undefined ||
      target.template.repository.url !==
        `https://github.com/${context.templateRepo.name}.git`
    ) {
      messages.push('chore(package): set template repo');

      target.template = {
        repository: {
          url: `https://github.com/${context.templateRepo.name}.git`
        }
      };
    }

    const toBeDeletedModules =
      target.devDependencies === undefined
        ? []
        : Array.from(
            context.optionalDevModules(
              new Set(Object.keys(target.devDependencies))
            )
          ).filter(m => !usedDevModules.has(m));

    toBeDeletedModules.forEach(d => {
      messages = messages.filter(
        m => !m.startsWith(`chore(devDependencies): add ${d}@`)
      );
      delete target.devDependencies[d];
    });

    target = deleter(target, template, messages, []);

    if (target.keywords !== undefined) {
      delete target.keywords['npm-package-template'];
    }

    if (
      template.template !== undefined &&
      template.template.keywords !== undefined
    ) {
      Object.keys(template.template.keywords).forEach(r =>
        addKeyword(
          target,
          new RegExp(r),
          template.template.keywords[r],
          messages
        )
      );
    }

    removeKeyword(target, ['null', null, undefined], messages);

    if (messages.length === 0) {
      messages.push('chore: update package.json from template');
    }

    let newContent = JSON.stringify(context.expand(target), undefined, 2);
    const lastChar = newContent[newContent.length - 1];

    // keep trailing newline
    if (originalLastChar === '\n' && lastChar === '}') {
      newContent += '\n';
    }

    return {
      content: newContent,
      messages,
      changed: original !== newContent
    };
  }
}

function deleter(object, reference, messages, path) {
  if (
    typeof object === 'string' ||
    object instanceof String ||
    object === true ||
    object === false ||
    object === undefined ||
    object === null ||
    typeof object === 'number' ||
    object instanceof Number
  ) {
    return object;
  }

  if (Array.isArray(object)) {
    return object.map((e, i) => {
      path.push(i);
      const n = deleter(
        object[i],
        Array.isArray(reference) ? reference[i] : undefined,
        messages,
        path
      );
      path.pop();
      return n;
    });
  }

  if (reference) {
    Object.keys(reference).forEach(key => {
      path.push(key);

      if (reference[key] === '--delete--' && object[key] !== undefined) {
        if (object[key] !== '--delete--') {
          messages.push(`chore(npm): delete ${path.join('.')}`);
        }
        delete object[key];
      } else {
        object[key] = deleter(object[key], reference[key], messages, path);
      }
      path.pop();
    });
  }

  return object;
}

function removeKeyword(pkg, keywords, messages) {
  if (pkg.keywords !== undefined) {
    keywords.forEach(keyword => {
      if (pkg.keywords.find(k => k === keyword)) {
        messages.push(`docs(package): remove keyword ${keyword}`);
        pkg.keywords = pkg.keywords.filter(k => k !== keyword);
      }
    });

    if (pkg.keywords[0] === null || pkg.keywords[0] === undefined) {
      messages.push(`docs(package): remove keyword null`);
      pkg.keywords = [];
    }
  }
}

function addKeyword(pkg, regex, keyword, messages) {
  if (keyword === undefined || keyword === null || keyword === 'null') {
    return;
  }

  if (pkg.name.match(regex)) {
    if (pkg.keywords === undefined) {
      pkg.keywords = [];
    }
    if (!pkg.keywords.find(k => k === keyword)) {
      messages.push(`docs(package): add keyword ${keyword}`);
      pkg.keywords.push(keyword);
    }
  }
}

const recast = require('recast');
//const types = require('ast-types');

class Rollup extends File {
  static matchesFileName(name) {
    return name.match(/rollup\.config\.js/);
  }

  get needsTemplate() {
    return false;
  }

  optionalDevModules(modules) {
    return new Set(
      Array.from(modules).filter(
        m => m.match(/rollup-plugin/) || m.match(/babel-preset/)
      )
    );
  }

  usedDevModules(content) {
    const modules = new Set();

    const ast = recast.parse(content);

    for (const decl of ast.program.body) {
      if (decl.type === 'ImportDeclaration') {
        modules.add(decl.source.value);
      }
    }

    return modules;
  }

  async mergeContent(context, original, template) {
    if (template === '') {
      return {
        content: original,
        changed: false
      };
    }
    if (original === '') {
      return {
        content: template,
        messages: ['chore(rollup): copy from template'],
        changed: true
      };
    }

    try {
      const templateAST = recast.parse(template);
      const ast = recast.parse(original);

      removeUseStrict(ast);

      const exp = exportDefaultDeclaration(ast);
      const templateExp = exportDefaultDeclaration(templateAST);

      if (exp !== undefined && exp.properties !== undefined) {
        let output, dest;

        const banner = removePropertiesKey(exp.properties, 'banner');

        for (const p of exp.properties) {
          switch (p.key.name) {
            case 'targets':
              p.key.name = 'output';
              dest = p.value.elements[0].properties[0]; //.find(x => x.name === 'dest');
              p.value = templateExp.properties.find(
                x => x.key.name === 'output'
              ).value;
              output = p;
              break;
            case 'entry':
              p.key.name = 'input';
              p.value = templateExp.properties.find(
                x => x.key.name === 'input'
              ).value;
          }
        }

        if (exp.properties.find(x => x.key.name === 'input') === undefined) {
          exp.properties.push(
            templateExp.properties.find(x => x.key.name === 'input')
          );
        }

        if (
          exp.properties.find(x => x && x.key.name === 'output') === undefined
        ) {
          output = templateExp.properties.find(x => x.key.name === 'output');
          exp.properties.push(output);
        }

        if (output !== undefined) {
          if (banner !== undefined) {
            output.value.properties.push(banner);
          }

          if (dest !== undefined) {
            output.value.properties.find(x => x.key.name === 'file').value =
              dest.value;
          }
        }

        removePropertiesKey(exp.properties, 'format');
        removePropertiesKey(exp.properties, 'sourceMap');
        removePropertiesKey(exp.properties, 'dest');
      }

      let pkg = importDeclaration(ast, 'pkg');
      if (pkg === undefined) {
        pkg = importDeclaration(templateAST, 'pkg');
        if (pkg !== undefined) {
          ast.program.body = [pkg, ...ast.program.body];
        }
      }

      const content = recast.print(ast).code;

      return {
        content,
        changed: content !== original,
        messages: ['chore(rollup): update from template']
      };
    } catch (e) {
      context.warn(`unable to parse ${this.path}`);
      context.warn(e);
    }

    return {
      content: original,
      changed: false
    };
  }
}

function removeUseStrict(ast) {
  for (const i in ast.program.body) {
    const decl = ast.program.body[i];
    if (
      decl.type === 'ExpressionStatement' &&
      decl.expression.type === 'Literal' &&
      decl.expression.value === 'use strict'
    ) {
      ast.program.body.splice(i, 1);
      return;
    }
  }
}

function importDeclaration(ast, localName) {
  for (const decl of ast.program.body) {
    if (
      decl.type === 'ImportDeclaration' &&
      decl.specifiers[0].local.name === localName
    ) {
      return decl;
    }
  }

  return undefined;
}

function exportDefaultDeclaration(ast) {
  for (const decl of ast.program.body) {
    if (decl.type === 'ExportDefaultDeclaration') {
      return decl.declaration;
    }
  }

  return undefined;
}

function removePropertiesKey(properties, name) {
  if (properties === undefined) {
    return undefined;
  }

  const toBeRemoved = properties.findIndex(x => x && x.key.name === name);
  if (toBeRemoved >= 0) {
    const slot = properties[toBeRemoved];
    properties.splice(toBeRemoved, 1);
    return slot;
  }

  return undefined;
}

class License extends File {
  static matchesFileName(name) {
    return name.match(/^LICENSE/);
  }

  async mergeContent(context, original, template) {
    const properties = context.properties;
    const messages = [];

    let year = properties['date.year'];

    const m = original.match(
      /opyright\s*\(c\)\s*(\d+)([,\-\d]+)*(\s*(,|by)\s*(.*))?/
    );

    if (m) {
      const years = new Set();
      years.add(parseInt(m[1], 10));

      if (m[2] !== undefined) {
        m[2].split(/\s*[,\-]\s*/).forEach(y => {
          if (y.length > 0) {
            years.add(parseInt(y, 10));
          }
        });
      }

      if (m[4] !== undefined) {
        properties['license.owner'] = m[5];
      }

      if (!years.has(year)) {
        years.add(year);
        messages.push(`chore(license): add current year ${year}`);
      }

      year = Array.from(years)
        .sort((a, b) => (a < b ? -1 : a > b ? 1 : 0))
        .join(',');
    }

    if (messages.length === 0) {
      messages.push('chore(license): update');
    }

    if (original !== '') {
      const content = original.replace(
        /opyright\s*\(c\)\s*(\d+)([,\-\d])*/,
        `opyright (c) ${year}`
      );

      return {
        changed: content !== original,
        messages,
        content
      };
    }

    return {
      content: context.expand(template),
      changed: true,
      messages: [`chore(license): add LICENSE`]
    };
  }
}

class MergeLineSet extends File {
  async mergeContent(context, original, template) {
    const result = new Set(template.split(/\n/));
    original.split(/\n/).forEach(line => result.add(line));

    const content = Array.from(result.values()).join('\n');

    return {
      content,
      changed: content !== original,
      messages: [this.options.message || 'fix: updated from template']
    };
  }
}

class MergeAndRemoveLineSet extends MergeLineSet {
  async mergeContent(context, original, template) {
    const toBeRemoved = new Set(
      template
        .split(/\n/)
        .filter(l => l.startsWith('- '))
        .map(l => l.replace(/^-\s+/, ''))
    );

    const result = new Set(template.split(/\n/).filter(l => !l.match(/^-\s+/)));
    original
      .split(/\n/)
      .filter(l => !toBeRemoved.has(l))
      .forEach(line => result.add(line));

    const content = Array.from(result.values()).join('\n');

    return {
      content,
      changed: content !== original,
      messages: [
        this.options.message || `chore: ${this.path} updated from template`
      ]
    };
  }
}

class ReplaceIfEmpty extends File {
  get needsTemplate() {
    return false;
  }

  async mergeContent(context, original, template) {
    return original === ''
      ? {
          content: context.expand(template),
          changed: template !== '',
          messages: [
            this.options.message ||
              `chore: add missing ${this.path} from template`
          ]
        }
      : {
          path: this.path,
          content: original,
          changed: false
        };
  }
}

class Replace extends File {
  async mergeContent(context, original, template) {
    const content = context.expand(template);

    return {
      content,
      changed: content !== original,
      messages: [`chore: ${this.path} overwritten from template`]
    };
  }
}

class JSONFile extends File {
  static matchesFileName(name) {
    return name.match(/\.json$/);
  }

  get needsTemplate() {
    return false;
  }

  async mergeContent(context, original, templateRaw) {
    if (templateRaw === '' || templateRaw === undefined) {
      return undefined;
    }

    const target =
      original === '' || original === undefined ? {} : JSON.parse(original);
    const template = JSON.parse(templateRaw);

    Object.assign(target, template);

    const content = JSON.stringify(context.expand(target), undefined, 2);

    return {
      content,
      changed: content !== original,
      messages: ['chore: update from template']
    };
  }
}

class JSDoc extends JSONFile {
  optionalDevModules(modules) {
    return new Set(Array.from(modules).filter(m => m.match(/babel-preset/)));
  }

  usedDevModules(content) {
    const modules = new Set();

    const json = JSON.parse(content);

    if (json.babel !== undefined && json.babel.presets !== undefined) {
      json.babel.presets.forEach(m => modules.add(`babel-preset-${m}`));
    }

    return modules;
  }
}

const mm = require('micromatch');

const mergers = [
  Rollup,
  Travis,
  Readme,
  Package,
  JSONFile,
  JSDoc,
  Travis,
  MergeAndRemoveLineSet,
  MergeLineSet,
  License,
  ReplaceIfEmpty,
  Replace
];

const defaultMapping = [
  { merger: 'Package', pattern: '**/package.json' },
  { merger: 'Travis', pattern: '.travis.yml' },
  { merger: 'Readme', pattern: '**/README.*' },
  { merger: 'JSDoc', pattern: '**/jsdoc.json' },
  { merger: 'Rollup', pattern: '**/rollup.config.js' },
  { merger: 'License', pattern: 'LICENSE' },
  {
    merger: 'MergeAndRemoveLineSet',
    pattern: '.gitignore',
    options: { message: 'chore(git): update {{path}} from template' }
  },
  {
    merger: 'MergeAndRemoveLineSet',
    pattern: '.npmignore',
    options: { message: 'chore(npm): update {{path}} from template' }
  },
  { merger: 'ReplaceIfEmpty', pattern: '**/*' }
];

async function createFiles(branch, mapping = defaultMapping) {
  const files = await branch.list();
  let alreadyPresent = new Set();

  return mapping
    .map(m => {
      const found = mm(
        files.filter(f => f.type === 'blob').map(f => f.path),
        m.pattern
      );

      const notAlreadyProcessed = found.filter(f => !alreadyPresent.has(f));

      alreadyPresent = new Set([...Array.from(alreadyPresent), ...found]);

      return notAlreadyProcessed.map(f => {
        const merger = mergers.find(merger => merger.name === m.merger);

        return new (merger ? merger : ReplaceIfEmpty)(f, m.options);
      });
    })
    .reduce((last, current) => Array.from([...last, ...current]), []);
}

async function npmTemplateSync(
  spinner,
  logger,
  token,
  targetRepo,
  templateRepo,
  dry = false
) {
  spinner.text = targetRepo;
  const [user, repo, branch = 'master'] = targetRepo.split(/[\/#]/);

  let newBranch;

  try {
    const provider = new githubRepositoryProvider.GithubProvider({ auth: token });
    const repository = await provider.repository(targetRepo);

    const maxBranchId = Array.from((await repository.branches()).keys()).reduce(
      (prev, current) => {
        const m = current.match(/template-sync-(\d+)/);
        if (m) {
          const r = parseInt(m[1], 10);
          if (r > prev) {
            return r;
          }
        }

        return prev;
      },
      0
    );

    const sourceBranch = await repository.branch(branch);
    const newBrachName = `template-sync-${maxBranchId + 1}`;

    const context = new Context(repository, undefined, {
      'github.user': user,
      'github.repo': repo,
      name: repo,
      user,
      'date.year': new Date().getFullYear(),
      'license.owner': user
    });

    context.logger = logger;
    context.dry = dry;
    context.spiner = spinner;

    const pkg = new Package('package.json');

    if (templateRepo === undefined) {
      templateRepo = await pkg.templateRepo(context);

      if (templateRepo === undefined) {
        throw new Error(
          `Unable to extract template repo url from ${targetRepo} ${pkg.path}`
        );
      }
    }

    context.templateRepo = await provider.repository(templateRepo);
    const templateBranch = await context.templateRepo.branch('master');

    const json = JSON.parse(
      await pkg.templateContent(context, { ignoreMissing: true })
    );

    const files = await createFiles(
      templateBranch,
      json.template && json.template.files
    );

    files.forEach(f => context.addFile(f));

    context.logger.debug(context.files.values());

    const merges = (await Promise.all(
      files.map(f => f.saveMerge(context, spinner))
    )).filter(m => m !== undefined && m.changed);

    if (merges.length === 0) {
      spinner.succeed(`${targetRepo}: nothing changed`);
      return;
    }

    spinner.text = merges.map(m => `${targetRepo}: ${m.messages[0]}`).join(',');

    if (dry) {
      spinner.succeed(`${targetRepo}: dry run`);
      return;
    }

    newBranch = await repository.createBranch(newBrachName, sourceBranch);

    const messages = merges.reduce((result, merge) => {
      merge.messages.forEach(m => result.push(m));
      return result;
    }, []);

    await newBranch.commit(messages.join('\n'), merges);

    try {
      const pullRequest = await sourceBranch.createPullRequest(newBranch, {
        title: `merge package template from ${context.templateRepo.name}`,
        body: merges
          .map(
            m =>
              `${m.path}
---
- ${m.messages.join('\n- ')}
`
          )
          .join('\n')
      });
      spinner.succeed(`${targetRepo}: ${pullRequest.name}`);

      return pullRequest;
    } catch (err) {
      if (newBranch !== undefined) {
        newBranch.delete();
      }

      spinner.fail(err);
    }
  } catch (err) {
    spinner.fail(`${user}/${repo}: ${err}`);
    throw err;
  }
}

exports.createFiles = createFiles;
exports.npmTemplateSync = npmTemplateSync;
